use super::challenge::Keys;
use super::InternalNodeUpdate;
use crate::{
    datastore::State,
    grpc::challenge::{update_client::UpdateClient, Empty},
};
use detee_sgx::RaTlsConfig;
use std::{str::FromStr, sync::Arc};
use tokio::{
    sync::broadcast::Sender,
    time::{sleep, Duration},
};
use tokio_stream::{wrappers::BroadcastStream, StreamExt};

#[derive(Clone)]
pub struct ConnManager {
    my_ip: String,
    state: Arc<State>,
    tx: Sender<InternalNodeUpdate>,
    ratls_config: RaTlsConfig,
}

impl ConnManager {
    pub fn init(
        my_ip: String,
        state: Arc<State>,
        ratls_config: RaTlsConfig,
        tx: Sender<InternalNodeUpdate>,
    ) -> Self {
        Self { my_ip, state, ratls_config, tx }
    }

    pub async fn start_with_node(self, node_ip: String) {
        self.connect_wrapper(node_ip).await;
    }

    pub async fn start(self) {
        loop {
            if let Some(ip) = self.state.get_random_disconnected_node() {
                println!("Found random disconnected node {ip}");
                self.connect_wrapper(ip.clone()).await;
            }
            sleep(Duration::from_secs(3)).await;
        }
    }

    async fn connect_wrapper(&self, node_ip: String) {
        let state = self.state.clone();
        state.add_conn(&node_ip);
        if let Err(e) = self.connect(node_ip.clone()).await {
            println!("Client connection for {node_ip} failed: {e:?}");
        }
        state.delete_conn(&node_ip);
    }

    async fn connect(&self, remote_ip: String) -> Result<(), Box<dyn std::error::Error>> {
        use detee_sgx::RaTlsConfigBuilder;
        use hyper::Uri;
        use hyper_util::{client::legacy::connect::HttpConnector, rt::TokioExecutor};
        use tokio_rustls::rustls::ClientConfig;

        println!("Connecting to {remote_ip}...");

        let tls = ClientConfig::from_ratls_config(self.ratls_config.clone())
            .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, format!("{}", e)))?;

        let mut http = HttpConnector::new();
        http.enforce_http(false);

        let cloned_node_ip = remote_ip.clone();

        let connector = tower::ServiceBuilder::new()
            .layer_fn(move |s| {
                let tls = tls.clone();

                hyper_rustls::HttpsConnectorBuilder::new()
                    .with_tls_config(tls)
                    .https_or_http()
                    .enable_http2()
                    .wrap_connector(s)
            })
            .map_request(move |_| {
                Uri::from_str(&format!("https://{cloned_node_ip}:31373"))
                    .expect("Could not parse URI")
            })
            .service(http);

        let client =
            hyper_util::client::legacy::Client::builder(TokioExecutor::new()).build(connector);

        let uri = Uri::from_static("https://example.com");
        let mut client = UpdateClient::with_origin(client, uri);

        let rx = self.tx.subscribe();
        let cloned_remote_ip = remote_ip.clone();
        let rx_stream =
            BroadcastStream::new(rx).filter_map(|n| n.ok()).filter_map(move |int_update| {
                if int_update.sender_ip != cloned_remote_ip {
                    Some(int_update.update)
                } else {
                    None
                }
            });

        let response = client.get_updates(rx_stream).await.map_err(|e| {
            println!("Error connecting to {remote_ip}: {e}");
            if e.to_string().contains("QuoteVerifyError") {
                self.state.increase_net_attacks();
            }
            e
        })?;
        let mut resp_stream = response.into_inner();

        // Immediately send our info as a network update
        let my_info = (self.my_ip.clone(), self.state.get_my_info()).into();
        let _ = self.tx.send(InternalNodeUpdate { sender_ip: self.my_ip.clone(), update: my_info });

        while let Some(update) = resp_stream.message().await? {
            // update the entire network in case the information is new
            if self.state.process_node_update(update.clone().into()) {
                // if process update returns true, the update must be forwarded
                if self
                    .tx
                    .send(InternalNodeUpdate {
                        sender_ip: remote_ip.clone(),
                        update: update.clone(),
                    })
                    .is_err()
                {
                    println!("Tokio broadcast receivers had an issue consuming the channel");
                };
            }
        }

        Ok(())
    }
}

pub async fn key_grabber(
    state: Arc<State>,
    node_ip: String,
    ratls_config: RaTlsConfig,
) -> Result<Keys, Box<dyn std::error::Error>> {
    use detee_sgx::RaTlsConfigBuilder;
    use hyper::Uri;
    use hyper_util::{client::legacy::connect::HttpConnector, rt::TokioExecutor};
    use tokio_rustls::rustls::ClientConfig;

    println!("Getting key from {node_ip}...");

    let tls = ClientConfig::from_ratls_config(ratls_config)
        .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, format!("{}", e)))?;

    let mut http = HttpConnector::new();
    http.enforce_http(false);

    let cloned_node_ip = node_ip.clone();

    let connector = tower::ServiceBuilder::new()
        .layer_fn(move |s| {
            let tls = tls.clone();

            hyper_rustls::HttpsConnectorBuilder::new()
                .with_tls_config(tls)
                .https_or_http()
                .enable_http2()
                .wrap_connector(s)
        })
        .map_request(move |_| {
            Uri::from_str(&format!("https://{cloned_node_ip}:31373")).expect("Could not parse URI")
        })
        .service(http);

    let client = hyper_util::client::legacy::Client::builder(TokioExecutor::new()).build(connector);

    let uri = Uri::from_static("https://example.com");
    let mut client = UpdateClient::with_origin(client, uri);

    let response = client.get_keys(tonic::Request::new(Empty {})).await.map_err(|e| {
        println!("Error getting keys from {node_ip}: {e}");
        if e.to_string().contains("QuoteVerifyError") {
            state.increase_net_attacks();
        }
        e
    })?;
    Ok(response.into_inner())
}
